@@ -113,6 +113,7 @@ struct _GVfsBackendFtp
 
   SoupAddress *		addr;
   char *		user;
+  gboolean              has_initial_user;
   char *		password;	/* password or NULL for anonymous */
 
   /* vfuncs */
@@ -179,12 +180,12 @@ ftp_connection_pop_job (FtpConnection *conn)
     {
       g_vfs_job_failed_from_error (conn->job, conn->error);
       g_clear_error (&conn->error);
-      result = TRUE;
+      result = FALSE;
     }
   else
     {
       g_vfs_job_succeeded (conn->job);
-      result = FALSE;
+      result = TRUE;
     }
 
   conn->job = NULL;
@@ -237,7 +238,8 @@ ftp_connection_set_error_from_response (FtpConnection *conn, guint response)
 	break;
       case 450: /* Requested file action not taken. File unavailable (e.g., file busy). */
       case 550: /* Requested action not taken. File unavailable (e.g., file not found, no access). */
-	/* FIXME: This is a lot of different errors */
+	/* FIXME: This is a lot of different errors. So we have to pretend to 
+	 * be smart here. */
 	code = G_IO_ERROR_NOT_FOUND;
 	msg = _("File unavailable");
 	break;
@@ -1054,7 +1056,8 @@ g_vfs_backend_ftp_push_connection (GVfsBackendFtp *ftp, FtpConnection *conn)
   if (conn == NULL)
     return;
 
-  ftp_connection_pop_job (conn);
+  if (conn->job)
+    ftp_connection_pop_job (conn);
 
   g_mutex_lock (ftp->mutex);
   if (ftp->queue)
@@ -1211,7 +1214,7 @@ do_mount (GVfsBackend *backend,
   char *username;
   char *password;
   char *display_name;
-  gboolean aborted, anonymous;
+  gboolean aborted, anonymous, break_on_fail;
   GPasswordSave password_save = G_PASSWORD_SAVE_NEVER;
   guint port;
 
@@ -1236,8 +1239,18 @@ do_mount (GVfsBackend *backend,
 	                    soup_address_get_name (ftp->addr),
 	                    port);
 
-  if (ftp->user &&
-      g_vfs_keyring_lookup_password (ftp->user,
+  username = NULL;
+  password = NULL;
+  break_on_fail = FALSE;
+  
+  if (ftp->user != NULL && strcmp (ftp->user, "anonymous") == 0)
+    {
+      anonymous = TRUE;
+      break_on_fail = TRUE;
+      goto try_login;
+    }
+  
+  if (g_vfs_keyring_lookup_password (ftp->user,
 				     soup_address_get_name (ftp->addr),
 				     NULL,
 				     "ftp",
@@ -1254,19 +1267,25 @@ do_mount (GVfsBackend *backend,
 
   while (TRUE)
     {
+      GAskPasswordFlags flags;
       if (prompt == NULL)
 	/* translators: %s here is the hostname */
 	prompt = g_strdup_printf (_("Enter password for ftp on %s"), host);
 
+      flags = G_ASK_PASSWORD_NEED_PASSWORD;
+        
+      if (!ftp->has_initial_user)
+        flags |= G_ASK_PASSWORD_NEED_USERNAME | G_ASK_PASSWORD_ANONYMOUS_SUPPORTED;
+      
+      if (g_vfs_keyring_is_available ())
+        flags |= G_ASK_PASSWORD_SAVING_SUPPORTED;
+      
       if (!g_mount_source_ask_password (
 			mount_source,
 		        prompt,
 			ftp->user,
 		        NULL,
-		        G_ASK_PASSWORD_NEED_USERNAME |
-		        G_ASK_PASSWORD_NEED_PASSWORD |
-		        G_ASK_PASSWORD_ANONYMOUS_SUPPORTED |
-		        (g_vfs_keyring_is_available () ? G_ASK_PASSWORD_SAVING_SUPPORTED : 0),
+                        flags,
 		        &aborted,
 		        &password,
 		        &username,
@@ -1280,6 +1299,13 @@ do_mount (GVfsBackend *backend,
 	  break;
 	}
 
+      /* NEED_USERNAME wasn't set */
+      if (ftp->has_initial_user)
+        {
+          g_free (username);
+          username = g_strdup (ftp->user);
+        }
+      
 try_login:
       g_free (ftp->user);
       g_free (ftp->password);
@@ -1296,12 +1322,16 @@ try_login:
 	}
       else
 	{
-	  ftp->user = username ? username : g_strdup ("");
-	  ftp->password = password;
+	  ftp->user = username ? g_strdup (username) : g_strdup ("");
+	  ftp->password = g_strdup (password);
 	  if (ftp_connection_login (conn, username, password) != 0)
 	    break;
 	}
-      if (!g_error_matches (conn->error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
+      g_free (username);
+      g_free (password);
+      
+      if (break_on_fail ||
+          !g_error_matches (conn->error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
 	break;
 
       g_clear_error (&conn->error);
@@ -1316,7 +1346,7 @@ try_login:
     }
   else
     {
-      if (prompt)
+      if (prompt && !anonymous)
 	{
 	  /* a prompt was created, so we have to save the password */
 	  g_vfs_keyring_save_password (ftp->user,
@@ -1340,13 +1370,13 @@ try_login:
 	  g_free (port_str);
 	}
 
+      if (ftp->has_initial_user)
+        g_mount_spec_set (mount_spec, "user", ftp->user);
+          
       if (g_str_equal (ftp->user, "anonymous"))
-	{
-	  display_name = g_strdup_printf (_("ftp on %s"), host);
-	}
+        display_name = g_strdup_printf (_("ftp on %s"), host);
       else
 	{
-	  g_mount_spec_set (mount_spec, "user", ftp->user);
 	  /* Translators: the first %s is the username, the second the host name */
 	  display_name = g_strdup_printf (_("ftp as %s on %s"), ftp->user, host);
 	}
@@ -1396,6 +1426,7 @@ try_mount (GVfsBackend *backend,
 
   ftp->addr = soup_address_new (host, port);
   ftp->user = g_strdup (g_mount_spec_get (mount_spec, "user"));
+  ftp->has_initial_user = ftp->user != NULL;
 
   return FALSE;
 }
@@ -1529,6 +1560,34 @@ do_start_write (GVfsBackendFtp *ftp,
     }
 }
 
+static void
+gvfs_backend_ftp_purge_cache_directory (GVfsBackendFtp *ftp,
+					const FtpFile * dir)
+{
+  g_static_rw_lock_writer_lock (&ftp->directory_cache_lock);
+  g_hash_table_remove (ftp->directory_cache, dir);
+  g_static_rw_lock_writer_unlock (&ftp->directory_cache_lock);
+}
+
+static void
+gvfs_backend_ftp_purge_cache_of_file (GVfsBackendFtp *ftp,
+				      FtpConnection * conn,
+				      const FtpFile * file)
+{
+  char *dirname, *filename;
+  FtpFile *dir;
+
+  filename = ftp_filename_to_gvfs_path (conn, file);
+  dirname = g_path_get_dirname (filename);
+  dir = ftp_filename_from_gvfs_path (conn, dirname);
+
+  gvfs_backend_ftp_purge_cache_directory (ftp, dir);
+
+  g_free (dir);
+  g_free (filename);
+  g_free (dirname);
+}
+
 /* forward declaration */
 static GFileInfo *
 create_file_info (GVfsBackendFtp *ftp, FtpConnection *conn, const char *filename, char **symlink);
@@ -1560,6 +1619,7 @@ do_create (GVfsBackend *backend,
     }
   file = ftp_filename_from_gvfs_path (conn, filename);
   do_start_write (ftp, conn, flags, "STOR %s", file);
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, file);
   g_free (file);
   return;
 
@@ -1583,6 +1643,7 @@ do_append (GVfsBackend *backend,
 
   file = ftp_filename_from_gvfs_path (conn, filename);
   do_start_write (ftp, conn, flags, "APPE %s", filename);
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, file);
   g_free (file);
   return;
 }
@@ -1615,6 +1676,7 @@ do_replace (GVfsBackend *backend,
 
   file = ftp_filename_from_gvfs_path (conn, filename);
   do_start_write (ftp, conn, flags, "STOR %s", file);
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, file);
   g_free (file);
   return;
 }
@@ -1984,47 +2046,54 @@ do_enumerate (GVfsBackend *backend,
 
   dir = ftp_filename_from_gvfs_path (conn, dirname);
   files = enumerate_directory (ftp, conn, dir, FALSE);
-  if (files != NULL)
+  if (ftp_connection_pop_job (conn))
     {
-      iter = ftp->dir_ops->iter_new (conn);
-      for (walk = files; walk; walk = walk->next)
+      ftp_connection_push_job (conn, G_VFS_JOB (job));
+      if (files != NULL)
 	{
-	  char *symlink = NULL;
-	  info = ftp->dir_ops->iter_process (iter,
-					     conn,
-					     dir,
-					     NULL,
-					     walk->data,
-					     query_flags & G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS ? NULL : &symlink);
-	  if (symlink)
+	  iter = ftp->dir_ops->iter_new (conn);
+	  for (walk = files; walk; walk = walk->next)
 	    {
-	      /* This is necessary due to our locking. 
-	       * And we must not unlock here because it might invalidate the list we iterate */
-	      symlink_targets = g_slist_prepend (symlink_targets, symlink);
-	      symlink_fileinfos = g_slist_prepend (symlink_fileinfos, info);
+	      char *symlink = NULL;
+	      info = ftp->dir_ops->iter_process (iter,
+						 conn,
+						 dir,
+						 NULL,
+						 walk->data,
+						 query_flags & G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS ? NULL : &symlink);
+	      if (symlink)
+		{
+		  /* This is necessary due to our locking. 
+		   * And we must not unlock here because it might invalidate the list we iterate */
+		  symlink_targets = g_slist_prepend (symlink_targets, symlink);
+		  symlink_fileinfos = g_slist_prepend (symlink_fileinfos, info);
+		}
+	      else if (info)
+		{
+		  g_vfs_job_enumerate_add_info (job, info);
+		  g_object_unref (info);
+		}
 	    }
-	  else if (info)
+	  ftp->dir_ops->iter_free (iter);
+	  g_static_rw_lock_reader_unlock (&ftp->directory_cache_lock);
+	  for (twalk = symlink_targets, fwalk = symlink_fileinfos; twalk; 
+	       twalk = twalk->next, fwalk = fwalk->next)
 	    {
+	      info = resolve_symlink (ftp, conn, fwalk->data, twalk->data);
+	      g_free (twalk->data);
 	      g_vfs_job_enumerate_add_info (job, info);
 	      g_object_unref (info);
 	    }
+	  g_slist_free (symlink_targets);
+	  g_slist_free (symlink_fileinfos);
 	}
-      ftp->dir_ops->iter_free (iter);
-      g_static_rw_lock_reader_unlock (&ftp->directory_cache_lock);
-      for (twalk = symlink_targets, fwalk = symlink_fileinfos; twalk; 
-	   twalk = twalk->next, fwalk = fwalk->next)
-	{
-	  info = resolve_symlink (ftp, conn, fwalk->data, twalk->data);
-	  g_free (twalk->data);
-	  g_vfs_job_enumerate_add_info (job, info);
-	  g_object_unref (info);
-	}
-      g_slist_free (symlink_targets);
-      g_slist_free (symlink_fileinfos);
-  
-      if (!ftp_connection_in_error (conn))
-	g_vfs_job_enumerate_done (G_VFS_JOB_ENUMERATE (conn->job));
+      
+      g_vfs_job_enumerate_done (job);
+      conn->job = NULL;
+      g_clear_error (&conn->error);
     }
+  else
+    g_assert (files == NULL);
 
   g_vfs_backend_ftp_push_connection (ftp, conn);
   g_free (dir);
@@ -2050,7 +2119,6 @@ do_set_display_name (GVfsBackend *backend,
   dir = ftp_filename_from_gvfs_path (conn, name);
   g_free (name);
   now = ftp_filename_construct (conn, dir, display_name);
-  g_free (dir);
   if (now == NULL)
     {
       g_set_error (&conn->error, 
@@ -2070,6 +2138,8 @@ do_set_display_name (GVfsBackend *backend,
   g_free (now);
   g_vfs_job_set_display_name_set_new_path (job, name);
   g_free (name);
+  gvfs_backend_ftp_purge_cache_directory (ftp, dir);
+  g_free (dir);
   g_vfs_backend_ftp_push_connection (ftp, conn);
 }
 
@@ -2094,10 +2164,31 @@ do_delete (GVfsBackend *backend,
 				  RESPONSE_PASS_500,
 				  "DELE %s", file);
   if (STATUS_GROUP (response) == 5)
-    ftp_connection_send (conn,
-			 0,
-			 "RMD %s", file);
+    {
+      response = ftp_connection_send (conn,
+				      RESPONSE_PASS_500,
+				      "RMD %s", file);
+      if (response == 550)
+	{
+	  const GList *files = enumerate_directory (ftp, conn, file, FALSE);
+	  if (files)
+	    {
+	      g_static_rw_lock_reader_unlock (&ftp->directory_cache_lock);
+	      g_set_error (&conn->error, 
+			   G_IO_ERROR,
+			   G_IO_ERROR_NOT_EMPTY,
+			   "%s", g_strerror (ENOTEMPTY));
+	    }
+	  else
+	    ftp_connection_set_error_from_response (conn, response);
+	}
+      else if (STATUS_GROUP (response) == 5)
+	{
+	  ftp_connection_set_error_from_response (conn, response);
+	}
+    }
 
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, file);
   g_free (file);
   g_vfs_backend_ftp_push_connection (ftp, conn);
 }
@@ -2121,6 +2212,7 @@ do_make_directory (GVfsBackend *backend,
 		       "MKD %s", file);
   /* FIXME: Compare created file with name from server result to be sure 
    * it's correct and otherwise fail. */
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, file);
   g_free (file);
 
   g_vfs_backend_ftp_push_connection (ftp, conn);
@@ -2198,6 +2290,8 @@ do_move (GVfsBackend *backend,
 		       0,
 		       "RNTO %s", destfile);
 
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, srcfile);
+  gvfs_backend_ftp_purge_cache_of_file (ftp, conn, destfile);
 out:
   g_free (srcfile);
   g_free (destfile);
