@@ -36,6 +36,9 @@
 #include "ghalvolume.h"
 #include "ghalmount.h"
 
+/* Protects all fields of GHalDrive that can change */
+G_LOCK_DEFINE_STATIC(hal_volume);
+
 struct _GHalVolume {
   GObject parent;
 
@@ -56,6 +59,7 @@ struct _GHalVolume {
   GFile *foreign_mount_root;
   GMount *foreign_mount;
   gboolean is_mountable;
+  gboolean ignore_automount;
 
   char *name;
   char *icon;
@@ -129,7 +133,20 @@ g_hal_volume_init (GHalVolume *hal_volume)
 {
 }
 
-const static struct {
+static gboolean
+changed_in_idle (gpointer data)
+{
+  GHalVolume *volume = data;
+  
+  g_signal_emit_by_name (volume, "changed");
+  if (volume->volume_monitor != NULL)
+    g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+  g_object_unref (volume);
+  
+  return FALSE;
+}
+
+static const struct {
   const char *disc_type;
   const char *icon_name;
   char *ui_name;
@@ -152,7 +169,7 @@ const static struct {
   {"hddvd_r",       "media-optical-hddvd-r", N_("HD DVD-R Disc"), N_("Blank HD DVD-R Disc")},
   {"hddvd_rw",      "media-optical-hddvd-rw", N_("HD DVD-RW Disc"), N_("Blank HD DVD-RW Disc")},
   {"mo",            "media-optical-mo", N_("MO Disc"), N_("Blank MO Disc")},
-  {NULL,            "media-optical", N_("Disc"), N_("Blank Disc")},
+  {NULL,            "media-optical", N_("Disc"), N_("Blank Disc")}
 };
 
 static const char *
@@ -167,9 +184,9 @@ get_disc_name (const char *disc_type, gboolean is_blank)
     }
   
   if (is_blank)
-    return disc_data[n].ui_name_blank;
+    return dgettext (GETTEXT_PACKAGE, disc_data[n].ui_name_blank);
   else
-    return disc_data[n].ui_name;
+    return dgettext (GETTEXT_PACKAGE, disc_data[n].ui_name);
 }
 
 
@@ -202,6 +219,36 @@ format_size_for_display (guint64 size)
   return str;
 }
 
+static char **
+dupv_and_uniqify (char **str_array)
+{
+  int n, m, o;
+  int len;
+  char **result;
+
+  result = g_strdupv (str_array);
+  len = g_strv_length (result);
+
+  for (n = 0; n < len; n++)
+    {
+      char *s = result[n];
+      for (m = n + 1; m < len; m++)
+        {
+          char *p = result[m];
+          if (strcmp (s, p) == 0)
+            {
+              for (o = m + 1; o < len; o++)
+                result[o - 1] = result[o];
+              len--;
+              result[len] = NULL;
+              m--;
+            }
+        }
+    }
+
+  return result;
+}
+
 static void
 do_update_from_hal (GHalVolume *mv)
 {
@@ -221,6 +268,8 @@ do_update_from_hal (GHalVolume *mv)
   HalDevice *drive;
   char *name;
   char *size;
+  gboolean is_crypto;
+  gboolean is_crypto_cleartext;
 
   volume = mv->device;
   drive = mv->drive_device;
@@ -238,6 +287,13 @@ do_update_from_hal (GHalVolume *mv)
   volume_fsusage = hal_device_get_property_string (volume, "volume.fsusage");
   volume_fstype = hal_device_get_property_string (volume, "volume.fstype");
 
+  is_crypto = FALSE;
+  is_crypto_cleartext = FALSE;
+  if (strcmp (hal_device_get_property_string (volume, "volume.fsusage"), "crypto") == 0)
+    is_crypto = TRUE;
+  if (strlen (hal_device_get_property_string (volume, "volume.crypto_luks.clear.backing_volume")) > 0)
+    is_crypto_cleartext = TRUE;
+
   if (volume_is_disc && volume_disc_has_audio && mv->foreign_mount_root != NULL)
     name = g_strdup (_("Audio Disc"));
   else
@@ -276,7 +332,11 @@ do_update_from_hal (GHalVolume *mv)
     }
 
   mv->name = name;
-  mv->icon = _drive_get_icon (drive); /* use the drive icon since we're unmounted */
+
+  if (is_crypto || is_crypto_cleartext)
+    mv->icon = g_strdup ("drive-encrypted");
+  else
+    mv->icon = _drive_get_icon (drive); /* use the drive icon since we're unmounted */
 
   if (hal_device_get_property_bool (volume, "volume.is_mounted"))
     mv->mount_path = g_strdup (hal_device_get_property_string (volume, "volume.mount_point"));
@@ -285,7 +345,7 @@ do_update_from_hal (GHalVolume *mv)
 
   g_object_set_data_full (G_OBJECT (mv), 
                           "hal-storage-device-capabilities",
-                          g_strdupv (hal_device_get_property_strlist (mv->drive_device, "info.capabilities")),
+                          dupv_and_uniqify (hal_device_get_property_strlist (mv->drive_device, "info.capabilities")),
                           (GDestroyNotify) g_strfreev);
 
   if (volume_disc_type != NULL && strlen (volume_disc_type) == 0)
@@ -296,33 +356,69 @@ do_update_from_hal (GHalVolume *mv)
                           (GDestroyNotify) g_free);
 }
 
-#ifdef _WITH_GPHOTO2
+#ifdef HAVE_GPHOTO2
 static void
 do_update_from_hal_for_camera (GHalVolume *v)
 {
   const char *vendor;
   const char *product;
+  const char *icon_from_hal;
+  const char *name_from_hal;
+  gboolean is_audio_player;
 
   vendor = hal_device_get_property_string (v->drive_device, "usb_device.vendor");
   product = hal_device_get_property_string (v->drive_device, "usb_device.product");
+  icon_from_hal = hal_device_get_property_string (v->device, "info.desktop.icon");
+  name_from_hal = hal_device_get_property_string (v->device, "info.desktop.name");
 
-  if (vendor == NULL)
+  is_audio_player = hal_device_has_capability (v->device, "portable_audio_player");
+
+  v->name = NULL;
+  if (strlen (name_from_hal) > 0)
+    v->name = g_strdup (name_from_hal);
+  else if (vendor == NULL)
     {
       if (product != NULL)
         v->name = g_strdup (product);
-      else
-        v->name = g_strdup (_("Camera"));
     }
   else
     {
       if (product != NULL)
         v->name = g_strdup_printf ("%s %s", vendor, product);
       else
-        v->name = g_strdup_printf (_("%s Camera"), vendor);
+        {
+          if (is_audio_player)
+            {
+              /* Translators: %s is the device vendor */
+              v->name = g_strdup_printf (_("%s Audio Player"), vendor);
+            }
+          else
+            {
+              /* Translators: %s is the device vendor */
+              v->name = g_strdup_printf (_("%s Camera"), vendor);
+            }
+        }
+    }
+  if (v->name == NULL)
+    {
+      if (is_audio_player)
+        v->name = g_strdup (_("Audio Player"));
+      else
+        v->name = g_strdup (_("Camera"));
     }
 
-  v->icon = g_strdup ("camera");
+  if (strlen (icon_from_hal) > 0)
+    v->icon = g_strdup (icon_from_hal);
+  else if (is_audio_player)
+    v->icon = g_strdup ("multimedia-player");
+  else
+    v->icon = g_strdup ("camera");
   v->mount_path = NULL;
+
+  g_object_set_data_full (G_OBJECT (v), 
+                          "hal-storage-device-capabilities",
+                          dupv_and_uniqify (hal_device_get_property_strlist (v->device, "info.capabilities")),
+                          (GDestroyNotify) g_strfreev);
 }
 #endif
 
@@ -333,6 +429,8 @@ update_from_hal (GHalVolume *mv, gboolean emit_changed)
   char *old_icon;
   char *old_mount_path;
 
+  G_LOCK (hal_volume);
+  
   old_name = g_strdup (mv->name);
   old_icon = g_strdup (mv->icon);
   old_mount_path = g_strdup (mv->mount_path);
@@ -340,8 +438,10 @@ update_from_hal (GHalVolume *mv, gboolean emit_changed)
   g_free (mv->name);
   g_free (mv->icon);
   g_free (mv->mount_path);
-#ifdef _WITH_GPHOTO2
-  if (hal_device_has_capability (mv->device, "camera"))
+#ifdef HAVE_GPHOTO2
+  if (hal_device_has_capability (mv->device, "camera") || 
+      (hal_device_has_capability (mv->device, "portable_audio_player") &&
+       hal_device_get_property_bool (mv->device, "camera.libgphoto2.support")))
     do_update_from_hal_for_camera (mv);
   else
     do_update_from_hal (mv);
@@ -365,15 +465,13 @@ update_from_hal (GHalVolume *mv, gboolean emit_changed)
            old_icon == NULL ||
            strcmp (old_name, mv->name) != 0 ||
            strcmp (old_icon, mv->icon) != 0))
-        {
-          g_signal_emit_by_name (mv, "changed");
-          if (mv->volume_monitor != NULL)
-            g_signal_emit_by_name (mv->volume_monitor, "volume_changed", mv);
-        }
+        g_idle_add (changed_in_idle, g_object_ref (mv));
     }
   g_free (old_name);
   g_free (old_icon);
   g_free (old_mount_path);
+
+  G_UNLOCK (hal_volume);
 }
 
 static void
@@ -419,7 +517,10 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
   HalDevice *drive_device;
   const char *storage_udi;
   const char *device_path;
-      
+  gboolean ignore_automount;
+
+  ignore_automount = FALSE;
+  
   if (hal_device_has_capability (device, "block"))
     {
       storage_udi = hal_device_get_property_string (device, "block.storage_device");
@@ -432,9 +533,12 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
       
       device_path = hal_device_get_property_string (device, "block.device");
     }
-#ifdef _WITH_GPHOTO2
-  else if (hal_device_has_capability (device, "camera"))
+#ifdef HAVE_GPHOTO2
+  else if (hal_device_has_capability (device, "camera") ||
+           (hal_device_has_capability (device, "portable_audio_player") &&
+            hal_device_get_property_bool (device, "camera.libgphoto2.support")))
     {
+
       /* OK, so we abuse storage_udi and drive_device for the USB main
        * device that holds this interface... 
        */
@@ -453,6 +557,10 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
 
       if (foreign_mount_root == NULL)
         return NULL;
+
+      /* We don't want to automount cameras as the gphoto backend
+         blocks access from other apps */
+      ignore_automount = TRUE;
     }
 #endif
   else
@@ -460,6 +568,11 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
       return NULL;
     }
 
+  if (drive_device && 
+      hal_device_has_property (drive_device, "storage.automount_enabled_hint") &&
+      !hal_device_get_property_bool (drive_device, "storage.automount_enabled_hint"))
+    ignore_automount = TRUE;
+  
   volume = g_object_new (G_TYPE_HAL_VOLUME, NULL);
   volume->volume_monitor = volume_monitor;
   g_object_add_weak_pointer (G_OBJECT (volume_monitor), (gpointer) &(volume->volume_monitor));
@@ -469,10 +582,11 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
   volume->drive_device = g_object_ref (drive_device);
   volume->foreign_mount_root = foreign_mount_root != NULL ? g_object_ref (foreign_mount_root) : NULL;
   volume->is_mountable = is_mountable;
+  volume->ignore_automount = ignore_automount || ! hal_device_is_recently_plugged_in (device);
   
   g_signal_connect_object (device, "hal_property_changed", (GCallback) hal_changed, volume, 0);
   g_signal_connect_object (drive_device, "hal_property_changed", (GCallback) hal_changed, volume, 0);
-  
+
   compute_uuid (volume);
   update_from_hal (volume, FALSE);
 
@@ -485,13 +599,16 @@ g_hal_volume_new (GVolumeMonitor   *volume_monitor,
 }
 
 /**
- * g_hal_volume_disconnected:
+ * g_hal_volume_removed:
  * @volume:
  * 
  **/
 void
 g_hal_volume_removed (GHalVolume *volume)
 {
+
+  G_LOCK (hal_volume);
+  
   if (volume->mount != NULL)
     {
       g_hal_mount_unset_volume (volume->mount, volume);
@@ -503,157 +620,210 @@ g_hal_volume_removed (GHalVolume *volume)
       g_hal_drive_unset_volume (volume->drive, volume);
       volume->drive = NULL;
     }
+
+  G_UNLOCK (hal_volume);
 }
 
 void
 g_hal_volume_set_mount (GHalVolume  *volume,
                         GHalMount *mount)
 {
-  if (volume->mount == mount)
-    return;
-  
-  if (volume->mount != NULL)
-    g_hal_mount_unset_volume (volume->mount, volume);
-  
-  volume->mount = mount;
-  
-  /* TODO: Emit changed in idle to avoid locking issues */
-  g_signal_emit_by_name (volume, "changed");
-  if (volume->volume_monitor != NULL)
-    g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+  G_LOCK (hal_volume);
+  if (volume->mount != mount)
+    {
+      
+      if (volume->mount != NULL)
+        g_hal_mount_unset_volume (volume->mount, volume);
+      
+      volume->mount = mount;
+
+      g_idle_add (changed_in_idle, g_object_ref (volume));
+    }
+  G_UNLOCK (hal_volume);
 }
  
 void
 g_hal_volume_unset_mount (GHalVolume  *volume,
                           GHalMount *mount)
 {
+  G_LOCK (hal_volume);
   if (volume->mount == mount)
     {
       volume->mount = NULL;
-      /* TODO: Emit changed in idle to avoid locking issues */
-      g_signal_emit_by_name (volume, "changed");
-      if (volume->volume_monitor != NULL)
-        g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+      g_idle_add (changed_in_idle, g_object_ref (volume));
     }
+  G_UNLOCK (hal_volume);
 }
 
 void
 g_hal_volume_set_drive (GHalVolume  *volume,
                         GHalDrive *drive)
 {
-  if (volume->drive == drive)
-    return;
-  
-  if (volume->drive != NULL)
-    g_hal_drive_unset_volume (volume->drive, volume);
-  
-  volume->drive = drive;
-  
-  /* TODO: Emit changed in idle to avoid locking issues */
-  g_signal_emit_by_name (volume, "changed");
-  if (volume->volume_monitor != NULL)
-    g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+  G_LOCK (hal_volume);
+  if (volume->drive != drive)
+    {
+      if (volume->drive != NULL)
+        g_hal_drive_unset_volume (volume->drive, volume);
+      
+      volume->drive = drive;
+      
+      g_idle_add (changed_in_idle, g_object_ref (volume));
+    }
+  G_UNLOCK (hal_volume);
 }
 
 void
 g_hal_volume_unset_drive (GHalVolume  *volume,
                           GHalDrive *drive)
 {
+  G_LOCK (hal_volume);
   if (volume->drive == drive)
     {
       volume->drive = NULL;
-      /* TODO: Emit changed in idle to avoid locking issues */
-      g_signal_emit_by_name (volume, "changed");
-      if (volume->volume_monitor != NULL)
-        g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+      g_idle_add (changed_in_idle, g_object_ref (volume));
     }
+  G_UNLOCK (hal_volume);
 }
 
 static GIcon *
 g_hal_volume_get_icon (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
-  return g_themed_icon_new_with_default_fallbacks (hal_volume->icon);
+  GIcon *icon;
+
+  G_LOCK (hal_volume);
+  icon = g_themed_icon_new_with_default_fallbacks (hal_volume->icon);
+  G_UNLOCK (hal_volume);
+  
+  return icon;
 }
 
 static char *
 g_hal_volume_get_name (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
-  return g_strdup (hal_volume->name);
+  char *name;
+
+  G_LOCK (hal_volume);
+  name = g_strdup (hal_volume->name);
+  G_UNLOCK (hal_volume);
+  
+  return name;
 }
 
 static char *
 g_hal_volume_get_uuid (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
-  return g_strdup (hal_volume->uuid);
+  char *uuid;
+
+  G_LOCK (hal_volume);
+  uuid = g_strdup (hal_volume->uuid);
+  G_UNLOCK (hal_volume);
+  
+  return uuid;
 }
 
 static gboolean
 g_hal_volume_can_mount (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
-  return hal_volume->is_mountable;
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = hal_volume->is_mountable;
+  G_UNLOCK (hal_volume);
+  
+  return res;
 }
 
 static gboolean
 g_hal_volume_can_eject (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = FALSE;
   if (hal_volume->drive != NULL)
-    return g_drive_can_eject (G_DRIVE (hal_volume->drive));
-  return FALSE;
+    res = g_drive_can_eject (G_DRIVE (hal_volume->drive));
+  G_UNLOCK (hal_volume);
+  
+  return res;
 }
 
 static gboolean
 g_hal_volume_should_automount (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
-  /* TODO: For now, just never automount things that are not
-     local. Need to figure out a better approach later. */
-  return hal_volume->foreign_mount == NULL;
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = ! (hal_volume->ignore_automount);
+  G_UNLOCK (hal_volume);
+  
+  return res;
 }
 
 static GDrive *
 g_hal_volume_get_drive (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  GDrive *drive;
+
+  G_LOCK (hal_volume);
+  drive = NULL;
   if (hal_volume->drive != NULL)
-    return g_object_ref (hal_volume->drive);
-  return NULL;
+    drive = g_object_ref (hal_volume->drive);
+  G_UNLOCK (hal_volume);
+  
+  return drive;
 }
 
 static GMount *
 g_hal_volume_get_mount (GVolume *volume)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  GMount *mount;
 
+  G_LOCK (hal_volume);
+  mount = NULL;
   if (hal_volume->foreign_mount != NULL)
-    return g_object_ref (hal_volume->foreign_mount);
+    mount = g_object_ref (hal_volume->foreign_mount);
+  else if (hal_volume->mount != NULL)
+    mount = g_object_ref (hal_volume->mount);
+  G_UNLOCK (hal_volume);
 
-  if (hal_volume->mount != NULL)
-    return g_object_ref (hal_volume->mount);
-
-  return NULL;
+  return mount;
 }
 
 gboolean
 g_hal_volume_has_mount_path (GHalVolume *volume,
                              const char  *mount_path)
 {
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = FALSE;
   if (volume->mount_path != NULL)
-    return strcmp (volume->mount_path, mount_path) == 0;
-  return FALSE;
+    res = strcmp (volume->mount_path, mount_path) == 0;
+  G_UNLOCK (hal_volume);
+
+  return res;
 }
 
 gboolean
 g_hal_volume_has_device_path (GHalVolume *volume,
                               const char  *device_path)
 {
+  gboolean res;
+
+  res = FALSE;
+  G_LOCK (hal_volume);
   if (volume->device_path != NULL)
-    return strcmp (volume->device_path, device_path) == 0;
-  return FALSE;
+    res = strcmp (volume->device_path, device_path) == 0;
+  G_UNLOCK (hal_volume);
+  return res;
 }
 
 gboolean
@@ -661,9 +831,14 @@ g_hal_volume_has_udi (GHalVolume  *volume,
                       const char  *udi)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
+  
+  G_LOCK (hal_volume);
+  res = FALSE;
   if (hal_volume->device != NULL)
-    return strcmp (hal_device_get_udi (hal_volume->device), udi) == 0;
-  return FALSE;
+    res = strcmp (hal_device_get_udi (hal_volume->device), udi) == 0;
+  G_UNLOCK (hal_volume);
+  return res;
 }
 
 gboolean
@@ -671,22 +846,34 @@ g_hal_volume_has_uuid (GHalVolume  *volume,
                        const char  *uuid)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = FALSE;
   if (hal_volume->uuid != NULL)
-    return strcmp (hal_volume->uuid, uuid) == 0;
-  return FALSE;
+    res = strcmp (hal_volume->uuid, uuid) == 0;
+  G_UNLOCK (hal_volume);
+
+  return res;
 }
 
 static void
 foreign_mount_unmounted (GMount *mount, gpointer user_data)
 {
   GHalVolume *volume = G_HAL_VOLUME (user_data);
-  if (volume->foreign_mount == mount)
+  gboolean check;
+
+  G_LOCK (hal_volume);
+  check = volume->foreign_mount == mount;
+  G_UNLOCK (hal_volume);
+  if (check)
     g_hal_volume_adopt_foreign_mount (volume, NULL);
 }
 
 void
 g_hal_volume_adopt_foreign_mount (GHalVolume *volume, GMount *foreign_mount)
 {
+  G_LOCK (hal_volume);
   if (volume->foreign_mount != NULL)
     g_object_unref (volume->foreign_mount);
 
@@ -698,9 +885,8 @@ g_hal_volume_adopt_foreign_mount (GHalVolume *volume, GMount *foreign_mount)
   else
     volume->foreign_mount =  NULL;
 
-  g_signal_emit_by_name (volume, "changed");
-  if (volume->volume_monitor != NULL)
-    g_signal_emit_by_name (volume->volume_monitor, "volume_changed", volume);
+  g_idle_add (changed_in_idle, g_object_ref (volume));
+  G_UNLOCK (hal_volume);
 }
 
 gboolean
@@ -708,9 +894,15 @@ g_hal_volume_has_foreign_mount_root (GHalVolume       *volume,
                                      GFile            *mount_root)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
+
+  G_LOCK (hal_volume);
+  res = FALSE;
   if (hal_volume->foreign_mount_root != NULL)
-    return g_file_equal (hal_volume->foreign_mount_root, mount_root);
-  return FALSE;
+    res = g_file_equal (hal_volume->foreign_mount_root, mount_root);
+  G_UNLOCK (hal_volume);
+  
+  return res;
 }
 
 
@@ -754,6 +946,7 @@ spawn_cb (GPid pid, gint status, gpointer user_data)
 
   g_simple_async_result_complete (simple);
   g_object_unref (simple);
+  g_object_unref (data->object);
   g_free (data);
 }
 
@@ -769,7 +962,7 @@ spawn_do (GVolume             *volume,
   GError *error;
 
   data = g_new0 (SpawnOp, 1);
-  data->object = G_OBJECT (volume);
+  data->object = g_object_ref (volume);
   data->callback = callback;
   data->user_data = user_data;
   data->cancellable = cancellable;
@@ -789,6 +982,7 @@ spawn_do (GVolume             *volume,
                                                      data->callback,
                                                      data->user_data,
                                                      error);
+      g_object_unref (data->object);
       g_simple_async_result_complete (simple);
       g_object_unref (simple);
       g_error_free (error);
@@ -831,6 +1025,7 @@ g_hal_volume_mount (GVolume             *volume,
               hal_volume->foreign_mount_root,
               hal_volume->device_path);*/
 
+  G_LOCK (hal_volume);
   if (hal_volume->foreign_mount_root != NULL)
     {
       ForeignMountOp *data;
@@ -856,6 +1051,7 @@ g_hal_volume_mount (GVolume             *volume,
         argv[4] = "-n";
       spawn_do (volume, cancellable, callback, user_data, argv);
     }
+  G_UNLOCK (hal_volume);
 }
 
 static gboolean
@@ -864,11 +1060,17 @@ g_hal_volume_mount_finish (GVolume       *volume,
                            GError       **error)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
 
-  if (hal_volume->foreign_mount_root != NULL)
-    return g_file_mount_enclosing_volume_finish (hal_volume->foreign_mount_root, result, error);
+  G_LOCK (hal_volume);
+  res = TRUE;
   
-  return TRUE;
+  if (hal_volume->foreign_mount_root != NULL)
+    res = g_file_mount_enclosing_volume_finish (hal_volume->foreign_mount_root, result, error);
+  
+  G_UNLOCK (hal_volume);
+  
+  return res;
 }
 
 typedef struct {
@@ -895,17 +1097,25 @@ g_hal_volume_eject (GVolume              *volume,
                     gpointer             user_data)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  GHalDrive *drive;
 
   /*g_warning ("hal_volume_eject");*/
 
+  drive = NULL;
+  G_LOCK (hal_volume);
   if (hal_volume->drive != NULL)
+    drive = g_object_ref (hal_volume->drive);
+  G_UNLOCK (hal_volume);
+  
+  if (drive != NULL)
     {
       EjectWrapperOp *data;
       data = g_new0 (EjectWrapperOp, 1);
-      data->object = G_OBJECT (volume);
+      data->object = g_object_ref (volume);
       data->callback = callback;
       data->user_data = user_data;
-      g_drive_eject (G_DRIVE (hal_volume->drive), flags, cancellable, eject_wrapper_callback, data);
+      g_drive_eject (G_DRIVE (drive), flags, cancellable, eject_wrapper_callback, data);
+      g_object_unref (drive);
     }
 }
 
@@ -915,10 +1125,14 @@ g_hal_volume_eject_finish (GVolume        *volume,
                           GError       **error)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  gboolean res;
 
+  G_LOCK (hal_volume);
+  res = TRUE;
   if (hal_volume->drive != NULL)
-    return g_drive_eject_finish (G_DRIVE (hal_volume->drive), result, error);
-  return TRUE;
+    res = g_drive_eject_finish (G_DRIVE (hal_volume->drive), result, error);
+  G_UNLOCK (hal_volume);
+  return res;
 }
 
 static char *
@@ -926,20 +1140,21 @@ g_hal_volume_get_identifier (GVolume              *volume,
                              const char          *kind)
 {
   GHalVolume *hal_volume = G_HAL_VOLUME (volume);
+  char *id;
 
+  G_LOCK (hal_volume);
+  id = NULL;
   if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_HAL_UDI) == 0)
-    return g_strdup (hal_device_get_udi (hal_volume->device));
+    id = g_strdup (hal_device_get_udi (hal_volume->device));
+  else if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE) == 0)
+    id = g_strdup (hal_volume->device_path);
+  else if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_LABEL) == 0)
+    id = g_strdup (hal_device_get_property_string (hal_volume->device, "volume.label"));
+  else if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_UUID) == 0)
+    id = g_strdup (hal_device_get_property_string (hal_volume->device, "volume.uuid"));
+  G_UNLOCK (hal_volume);
   
-  if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE) == 0)
-    return g_strdup (hal_volume->device_path);
-  
-  if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_LABEL) == 0)
-    return g_strdup (hal_device_get_property_string (hal_volume->device, "volume.label"));
-  
-  if (strcmp (kind, G_VOLUME_IDENTIFIER_KIND_UUID) == 0)
-    return g_strdup (hal_device_get_property_string (hal_volume->device, "volume.uuid"));
-  
-  return NULL;
+  return id;
 }
 
 static char **
@@ -949,6 +1164,8 @@ g_hal_volume_enumerate_identifiers (GVolume *volume)
   GPtrArray *res;
   const char *label, *uuid;
 
+  G_LOCK (hal_volume);
+  
   res = g_ptr_array_new ();
 
   g_ptr_array_add (res,
@@ -971,6 +1188,8 @@ g_hal_volume_enumerate_identifiers (GVolume *volume)
 
   /* Null-terminate */
   g_ptr_array_add (res, NULL);
+
+  G_UNLOCK (hal_volume);
   
   return (char **)g_ptr_array_free (res, FALSE);
 }
